
import geopandas as gpd
import ee
ee.Initialize()
import pandas as pd
from time import time
import os
import warnings
warnings.filterwarnings("ignore")

from waterpyk import default_saving_dir, in_colab_shell # Determine default saving behavior
from waterpyk import plots, watershed, gee, calcs

class StudyArea:
        
    def get_kind(self):
        """
        Adds attribute of point or watershed to object.
        
        """
        if len(self.coords) >1:
            self.kind = 'point'
        else: self.kind = 'watershed'
        return self
    
    def get_location(self, **kwargs):
        """
        Convert coordinates to GEE feature. For a USGS watershed, coordinates are extracted from USGS website using gage ID.

        Args:
            site_name (str, optional): Watersheds are already named. Default for lat/long sites is empty string.
       
        Returns:
            :obj:`feature`: gee feature with location geometry 
        """
        # Make sure the self.kind exists
        StudyArea.get_kind(self)
        
        # Get the GEE and geopandas geometries and metadata for a point
        if self.kind == 'point':
            lat = self.coords[0]
            long = self.coords[1]
            site_name = kwargs['site_name']
            description = 'Site at coordinates ' + str(lat) + ', ' + str(long) + '. Name = ' + site_name + '. CRS = EPSG:4326.'
            gee_feature = ee.Feature(ee.Geometry.Point(long, lat))
            gp_site = pd.DataFrame({'longitude': [long], 'latitude': [lat]})
            gpd_geometry = gpd.points_from_xy(gp_site.longitude, gp_site.latitude, crs="EPSG:4326")

        # Use the watershed package to get the geometries and metadata for a USGS gauged watershed
        elif self.kind == 'watershed':
            gage = str(self.coords[0])
            site_name, description = watershed.extract_metadata(gage)
            gee_feature, gpd_geometry = watershed.extract_geometry(gage)
        
        self.site_name = site_name
        self.description = description
        self.gee_feature = gee_feature
        self.gpd_geometry = gpd_geometry
    
        return self, gee_feature
    
            
    def get_data(self, layers, **kwargs):
        """
        Updates self with attributes containing dataframes for the site.
        If data already exists in saving_dir, data is simply loaded.
        Otherwise, data from layers is extracted from GEE and the USGS.
        
        Args:
            layers (str or :obj:`df`, optional): If str, specify 'minimal' or 'all' to extract default set of assets. If df, columns that must be present include: asset_id, start_date, end_date, relative_date, scale, bands, bands_to_scale, new_bandnames, scaling factor. These are the same parameters required for extract_basic(). 
            **interp (bool, optional): (default: True), currently no option to change to False.
            **combine_ET_bands (bool, optional): (default True) add ET bands to make one ET band.
            **bands_to_combine (list of str, optional): (default [Es, Ec]) ET bands to combine
            **band_names_combined (str, optional): (default 'ET') name of combined ET band
            **et_asset (str, optional): asset to be used for creation of ET column (default 'pml').
            **et_band (str, optional): band name to be used for creation of ET column (default 'ET').
            **ppt_asset (str, optional): asset name for precipitation column (default 'prism').
            **ppt_band (str, optional): band name for precipitation colum (default 'ppt')
            **snow_asset (str, optional): defaults to 'modis_snow'
            **snow_band (str, optional): defaults to 'snow'. Note: You will need to change this if you don't specify to change the default name of this asset upon extraction.
            **snow_correction (bool, optional): (default True) use snow correction factor when calculating deficit
            **snow_frac (int, optional): (default 10) set all ET when snow is greater than this (%) to 0 if snow_correction = True

        """
        if in_colab_shell(): path_format = '_'
        else: path_format = '/'
        
        ### If data already exists, just pull it in:
        if os.path.exists(self.saving_path):
            print('\nRetrieving site data from ' + os.path.abspath(self.saving_path))
            self.daily_df_long = pd.read_csv(self.saving_path + path_format + 'daily_df_long.csv')
            self.daily_df_wide = pd.read_csv(self.saving_path + path_format + 'daily_df_wide.csv')
            self.stats = pd.read_csv(self.saving_path + path_format + 'stats.csv')
            self.streamflow = pd.read_csv(self.saving_path + path_format + 'streamflow.csv')
            self.deficit_timeseries = pd.read_csv(self.saving_path + path_format + 'deficit_timeseries.csv')
            self.wateryear_totals = pd.read_csv(self.saving_path + path_format + 'wateryear_totals.csv')         
            self.smax = round(self.deficit_timeseries.D.max())
            self.maxdmax = round(self.deficit_timeseries.D_wy.max())
            self.deficit_timerange = (self.deficit_timeseries.date.min(), self.deficit_timeseries.date.max())
        
        ### Otherwise, use layers.csv to download new data and save it:
        else:
            if layers is None:
                print('No layers were specified for extraction and data is not already available.')
            try:
                df_long, df_image = gee.extract(layers, self.gee_feature, self.kind, **kwargs)
            except AttributeError:
                StudyArea.get_location(self, **kwargs)
                df_long, df_image = gee.extract(layers, self.gee_feature, self.kind, **kwargs)
        
            ### Convert dataframe to wide format using **kwargs
            df_wide = calcs.make_wide_df(df_long, **kwargs)

            ### If deficit data types are given, merge deficit data
            df_deficit = calcs.deficit(df_long, df_wide, **kwargs)
            df_deficit_min = df_deficit.date.min()
            df_deficit_max = df_deficit.date.max()
            df_wide = calcs.merge(df_wide, df_deficit, 'deficit')

            ### If kind = watershed, get and merge streamflow data
            if self.kind == 'watershed':
                gage = self.coords[0]
                df_streamflow = watershed.extract_streamflow(gage, **kwargs)
                df_wide = calcs.merge(df_wide, df_streamflow, 'streamflow')
            else: df_streamflow = pd.DataFrame()
            
            ### Create wateryear cumulative and total dataframes
            df_wide, df_total = calcs.wateryear(df_wide)

            ### Save data (just long df for GoogleColab and all data otherwise)
            print("\nSaving all dataframes at:\n\t% s" % os.path.abspath(self.saving_path))
            if in_colab_shell() == False: os.mkdir(self.saving_path)
            with open(self.saving_path + path_format + 'daily_df_long.csv', 'w') as f:
                df_long.to_csv(f)
            with open(self.saving_path + path_format + 'daily_df_wide.csv', 'w') as f:
                df_wide.to_csv(f)
            with open(self.saving_path + path_format + 'streamflow.csv', 'w') as f:
                df_streamflow.to_csv(f)
            with open(self.saving_path + path_format + 'deficit_timeseries.csv', 'w') as f:
                df_deficit.to_csv(f)    
            with open(self.saving_path + path_format + 'wateryear_totals.csv', 'w') as f:
                df_total.to_csv(f)    
            with open(self.saving_path + path_format + 'stats.csv', 'w') as f:
                df_image.to_csv(f)    
                    
            ### Add all attributes to self
            self.daily_df_long = df_long
            self.daily_df_wide = df_wide
            self.stats = df_image
            self.streamflow = df_streamflow
            self.deficit_timeseries = df_deficit
            self.wateryear_totals = df_total
            self.smax = round(df_deficit.D.max())
            self.maxdmax = round(df_deficit.D_wy.max())
            self.deficit_timerange = (df_deficit_min, df_deficit_max)
            
        
    def describe(self):
        """
        Print statements describing StudyArea attributes and deficit parameters, if deficit was calculated.

        """
        print('\n'+ str(self.description))
        print('Geometry kind:', self.kind)
        print('Data extracted from GEE:', self.daily_df_long.band.unique())
        print('GEE reducer used: MEAN() for watersheds and FIRST() for points')
        print('Data available for wateryear totals:', list(self.wateryear_totals))
        print('Deficit results:\n\tSmax = ' + str(self.smax) + ' mm')
        print('\tmax(Dmax) = ' + str(self.maxdmax) + ' mm')
        print('\tStart date: ' + str(self.deficit_timerange[0]) + '\n\tEnd date: ' + str(self.deficit_timerange[1]))
        print('Kwargs set to:', self.settings)
        
        
    def plot(self, kind, **kwargs):
        """ Make common plots of the data specifying what 'kind' of plot.
        Each kind of plot has defaults pre-set, which may differ from the arg defaults specified below.
        To see the full list of defaults, see the plots submodule.
        In general, updating kwargs should not be necessary except for small details like legend, title, or colors.
        
        Args:
            kind (str): The kind of plot you wanted. Not to be confused with the kind of site (ie watershed or point). Supported types are: 'timeseries', 'spearman', 'wateryear', 'RWS', and 'distribution'.
            **plot_PET (bool, optional)
            **plot_Q (bool, optional)
            **plot_D (bool, optional)
            **plot_Dwy (bool, optional):
            **plot_ET (bool, optional): include wateryear ET in plot
            **plot_ET_dry  (bool, optional): include dry season ET in plot
            **color_PET (str, optional): hex code string for color
            **color_Q (str, optional): hex code string for color
            **color_P (str, optional): hex code string for color
            **color_D (str, optional): hex code string for color
            **color_ET (str, optional): hex code string for color
            **color_WY (str, optional): hex code string for color
            **markeredgecolor (str, optional): default = black
            **linestyle_PET (str, optional)
            **linestyle_Q (str, optional)
            **linestyle_P (str, optional)
            **linestyle_D (str, optional)
            **linestyle_Dwy (str, optional)
            **linestyle_ET (str, optional)
            **lw (float, optional): line weight. Default = 1.5
            **xmin (date or float or int, optional)
            **xmax (date or float or int, optional)
            **legend (bool, optional): default = True
            **dpi (int, optional): default = 300
            **figsize (tuple, optional): default = (6,4)
            **xlabel (str, optional)
            **ylabel (str, optional)
            **twinx (bool, optional): some plots allow for twin x-axes. Default = False.
            **title (str, optional): default None

        """
        if kind == 'timeseries':
            fig = plots.plot_timeseries(self, **kwargs)
        elif kind == 'spearman':
            fig = plots.plot_spearman(self, **kwargs)
        elif kind == 'wateryear':
            fig = plots.plot_wateryear_totals(self, **kwargs)
        elif kind == 'RWS':
            fig = plots.plot_RWS(self, **kwargs)
        elif kind == 'distribution':
            fig = plots.plot_p_distribution(self, **kwargs)
        return fig
            
            
    def _path(self):
        """Make a folder name for storing data."""
        if self.kind == 'watershed':
            folder_name = str(self.coords[0])
        elif self.kind == 'point':
            folder_name = str(self.coords[0]) + '_' + str(self.coords[1])
        saving_path = os.path.abspath(os.path.join(self.saving_dir, folder_name))
        self.saving_path = saving_path
        return saving_path   
        
        
    def __init__(self, coords, layers = None, saving_dir = default_saving_dir, **kwargs):
        default_kwargs = {
            'site_name':'',
            'interp': True,
            'combine_ET_bands': True,
            'bands_to_combine': ['Es', 'Ec'],
            'band_name_final': 'ET',
            'et_asset': 'pml',
            'et_band': 'ET',
            'ppt_asset': 'prism',
            'ppt_band': 'ppt',
            'snow_asset':'modis_snow',
            'snow_band':'snow',
            'snow_correction': True,
            'snow_frac': 10,
            'flow_start_date':'1980-10-01',
            'flow_end_date':'2021-10-01'
            }
        kwargs = {**default_kwargs, **kwargs}
        self.settings = kwargs
        self.coords = coords
        self.layers = layers
        self.saving_dir = saving_dir
        self.get_kind()
        self._path()
        self.get_location(**kwargs)
        t1 = time()
        self.get_data(layers, **kwargs)
        t2 = time()
        print('\nTime to access data: ' + str(round(t2-t1,3)) + ' seconds')
        self.MAP = round(self.wateryear_totals.P.mean())


import pandas as pd
import numpy as np
import datetime

def interp_daily(df):
    """Interpolate all data to daily.
        
        Args:
            :obj:`df`: initial long-form dataframe for a single asset (may contain multiple bands) with column 'value' for interpolating
        
        Returns:
           :obj:`df`: dataframe with 'value' column containing daily interpolated daily 
    """
    df_interp = pd.DataFrame()
    temp = pd.DataFrame()
    diff_days = (df.date.max() - df.date.min()).days
    temp['date'] = [df.date.min() + datetime.timedelta(days=x) for x in range(0, diff_days)]
    for i in df.band.unique():
        df_temp = df[df['band']==i]
        df_temp = df_temp.merge(temp, how = 'right', on = 'date')
        df_temp['value'] = df_temp['value'].interpolate(method = "linear")
        df_temp['band'] = i
        df_interp = pd.concat([df_interp,df_temp])
    return df_interp

def combine_bands(df, bands_to_combine, band_name_final):
    """
    Defaults to add together soil and vegetation ET bands (Es and Ec) for PML to create ET band.
    Can be customized to add any bands together to create a new band.
    
    Args:
        bands_to_combine (list of str): list of strings of band names (example: ['Es', 'Ec']) to combine into a single band
        band_name_final (str): the final band name to be stored in the 'band' column of the final dataframe, where the 'value' column will contain the added values from bands_to_combine
    
    Returns:
       :obj:`df`: long-form dataframe from input with appended section which contains date, asset_name, value, band, and value_raw (set to np.nan) columns with data from combined
    """
    to_store = pd.DataFrame()
    for i in bands_to_combine:
        if i == bands_to_combine[0]:
            to_store = df[df['band'] == i]
            to_store.loc[:,'value_raw'] = np.nan
        else:
            subset = df[df['band']==i][['date', 'value']]
            to_store = to_store.merge(subset, how = 'inner', on = 'date')
    val_cols = to_store.loc[:,to_store.columns.str.startswith("value")]
    to_store['value'] = val_cols.sum(axis=1)
    to_store['band'] = band_name_final
    to_store = to_store[['date', 'asset_name', 'value', 'band', 'value_raw']]
    df = pd.concat([df, to_store])
    #df = df.append(to_store)
    df = df.reset_index(drop=True)
    return df



def make_wide_df(df_long, **kwargs):
    """
    Uses ET and P asset and band names (designated in **kwargs) to return a wide-form dataframe with columns:
    date, ET, Ei, P, and P-Ei (where Ei is the interception band from PML and is only extracted if it exists). merge_wide_streamflow() can then be used to merge with streamflow df. 
    
    Args:
        df_long: dataframe such as that produced by extract_assets_at_site() where columns contain asset_name, band, and value.
        **et_asset (str, optional): asset to be used for creation of ET column (default 'pml').
        **et_band (str, optional): band name to be used for creation of ET column (default 'ET').
        **ppt_asset (str, optional): asset name for precipitation column (default 'prism').
        **ppt_band (str, optional): band name for precipitation colum (default 'ppt')
        
    Returns:
        :obj:`df`: wide-form dataframe with columns for date, ET, Ei, P, and P-Ei (where Ei is the interception band from PML and is only included if it exists.)
    """

    default_kwargs = {
        'et_asset': 'pml',
        'et_band': 'ET',
        'ppt_asset': 'prism',
        'ppt_band': 'ppt',
    }
    kwargs = {**default_kwargs, **kwargs} 
        
    # Isolate ET dataset
    et_df = df_long[df_long['asset_name'] == kwargs['et_asset']]
    et_df = et_df[et_df['band'] == kwargs['et_band']]
    et_df['ET'] = et_df['value']
    
    # Isolate P dataset
    ppt_df = df_long[df_long['asset_name'] == kwargs['ppt_asset']]
    ppt_df = ppt_df[ppt_df['band'] == kwargs['ppt_band']]
    ppt_df['P'] = ppt_df['value']
    
    # Merge P + Ei (interception) if Ei band exists
    if 'Ei' in et_df.band.unique():
        ei_df = et_df[et_df['band'] == 'Ei']
        ei_df['Ei'] = ei_df['value']
        ppt_df = ppt_df.merge(ei_df[['date','Ei']], how = 'left', on = 'date')
        ppt_df['P_min_Ei'] = ppt_df['P'] - ppt_df['Ei']
    
        # Merge ET and P
        df_wide = et_df.merge(ppt_df, how = 'inner', on = 'date')[['date', 'ET', 'P', 'P_min_Ei', 'Ei']]
    
    else:
        # Merge ET and P
        df_wide = et_df.merge(ppt_df, how = 'inner', on = 'date')[['date', 'ET', 'P']]
        
    # Add wateryear column
    df_wide = df_wide.set_index(pd.to_datetime(df_wide['date']))
    df_wide['wateryear'] = np.where(~df_wide.index.month.isin([10,11,12]),df_wide.index.year,df_wide.index.year+1)
    df_wide = df_wide.reset_index(drop=True)
    
    return df_wide


def merge(df_wide, df_merge, merge_with, column_names = None):
    """Merge df_wide (dataframe with columns representing variable names) with streamflow dataframe.
   
    Args:
        df_wide :obj:`df`): wide-form dataframe with 'date' column and other columns with variable names (ie ET, P, etc)
        df_merge (:obj:`df`): dataframe with stuff to merge (streamflow or deficit). Must contain columns for 'date' and column_names if given.
        merge_with (str): choice of 'streamflow' or 'deficit' to merge with df_wide.
        column_names (list of str): Names of the column in df_merge that you wish to merge (default for streamflow: ['Q_mm'], default for deficit: ['D', and 'D_wy']). 'date' is automatically included.
    
    Returns:
        :obj:`df`: merged dataframes.
    """
    # Make sure they are both datetime format
    df_wide['date'] = pd.to_datetime(df_wide['date'])
    df_merge['date'] = pd.to_datetime(df_merge['date'])

    if column_names is None: 
        if merge_with == 'streamflow':
            column_names = ['date','Q_mm']
        elif merge_with == 'deficit':
            column_names = ['date', 'D', 'D_wy']
        else:
            print('merge() works for deficit or streamflow dataframes only.')
    elif 'date' in column_names: pass
    else: column_names = column_names + ['date']
    df_wide = df_wide.merge(df_merge[column_names], how = 'left', on = 'date')
    return df_wide


def wateryear(df_wide):
    """Create total and cumulative wateryear dataframes for all variables (ie columns) given in df_wide.
    
    Args:
        df_wide (:obj:`df`): wide-form dataframe with 'date' column and other columns with variable names (i.e. ET, P, Q_mm, etc.) This dataframe can also have D, D_wy, Q_mm, etc merged into it.
    
    Returns:
        :obj:`df`, :obj:`df`: 2 datafrmes: (1) the  original wide-form dataframe with added columns (in the naming form of ORIGINALNAME_cumulative) with cumulative wateryear values. For example, if 'P' was in original dataframe, then 'P_cumulative' now exists. If 'Q_mm' was in dataframe, then 'Q_mm_cumulative' now exists. (2) dataframe with one row for each wateryear with columns such as wateryear, ET, P, D_wy_max, ET_summer, Q_mm, etc, which represent the total (or maximum, for D_wy_max) wateryear sum.
    """
     # Create wateryear total dataframe and fill in wateryear and Dwy_max if exists
    df_total = pd.DataFrame()
    df_total['wateryear'] = df_wide.groupby(['wateryear'])['wateryear'].first()
    if 'D_wy' in df_wide:
        df_total['D_wy_max'] = df_wide.groupby(['wateryear'])['D_wy'].max()
        
    # Add summer ET to total wateryear df
    df_summer = df_wide.copy()
    df_summer = df_summer.set_index(df_summer['date'])
    df_summer['season'] = np.where(~df_summer.index.month.isin([6,7,8,9]),'summer','other')
    df_summer = df_summer[df_summer['season'] == 'summer']
    df_total['ET_summer'] = df_summer.groupby(['wateryear'])['ET'].sum()

    # Add wateryear cumulative to df_wide
    cols = ['ET', 'P', 'Ei', 'P_min_Ei', 'Q_mm']
    for col in cols:
        if col in df_wide:
            df_wide[col + '_cumulative'] = df_wide.groupby(['wateryear'])[col].cumsum()
            df_total[col] = df_wide.groupby(['wateryear'])[col].sum()

    # Calculate dV if all data is present
    if 'P' and 'ET' and 'Q_mm' in df_wide:
        df_wide['dV'] = df_wide['P_cumulative'] - df_wide['ET_cumulative'] - df_wide['Q_mm_cumulative'] 
        
    return df_wide, df_total
    

def deficit(df_long, df_wide = None, **kwargs):
    """
    Calculate D(t) after McCormick et al., 2021 and Dralle et al., 2020.

    Args:
        df_long (:obj:`df`): original long-style dataframe with snow, P, and ET data at a minimum
        df_wide (:obj:`df`): dataframe created from make_wide_df(). (default = None, in which case new df_wide is created from df_long using default args). This dataframe must have ET, P, and wateryear columns.
        **interp: (bool, optional): default: True. Currently no option to change to False.
        **combine_ET_bands (bool, optional): (default True) add ET bands to make one ET band.
        **bands_to_combine (list of str, optional): (default [Es, Ec]) ET bands to combine
        **band_names_combined (str, optional): (default ET) name of combined ET band
        **et_asset (str, optional): (default pml) ET dataset to use for deficit calculation, if multiple are given
        **et_band (str, optional): (default ET) band from ET dataset to use for deficit calculation, if multiple are given
        **ppt_asset (str, optional): (default prism) precipitation dataset to use for deficit calculation, if multiple are given
        **ppt_band (str, optional): (default ppt) precipitation dataset to use for deficit calculation, if multiple are given
        **snow_band (str, optional): defaults to 'snow'. Note: You will need to change this if you don't specify to change the default name of this asset upon extraction.
        **snow_correction (bool, optional): (default True) use snow correction factor when calculating deficit
        **snow_correction (bool, optional): (default True) use snow correction factor when calculating deficit
        **snow_frac (int, optional): (default 10) set all ET when snow is greater than this (%) to 0 if snow_correction = True

    Returns:
        :obj:`df`: dataframe with root-zone water storage deficit data where deficit is column 'D' and wateryear deficit is 'D_wy'.
    """
    
    default_kwargs = {
            'et_asset': 'pml',
            'et_band': 'ET',
            'ppt_asset': 'prism',
            'ppt_band': 'ppt',
            'snow_asset':'modis_snow',
            'snow_band':'snow',
            'snow_correction': True,
            'snow_frac': 10,
        }
    kwargs = {**default_kwargs, **kwargs} 
    
    # Make deficit dataframe if none given (just normal wide dataframe but only keep relevant data)
    if df_wide is None:
        df_deficit = make_wide_df(df_long, pivot_all = False, **kwargs)
    else: df_deficit = df_wide.copy()
    
    # Check to make sure everything we need is here
    necessary_columns = ['date', 'ET', 'P', 'wateryear'] 
    for col in necessary_columns:
        if col not in df_deficit:
            raise Exception(col + ' missing. Deficit cannot be calculated. Check assets specified in layers.')
    
    df_deficit = df_deficit[necessary_columns]

    # If using snow_correction, add snow to df_wide
    if kwargs['snow_correction'] == True:
        try:  
            snow_df = df_long[df_long['asset_name'] == kwargs['snow_asset']]
            snow_df = snow_df[snow_df['band'] == kwargs['snow_band']]
            snow_df['Snow'] = snow_df['value']
            # Merge snow together and filter ET by snow fraction
            df_deficit = df_deficit.merge(snow_df, how = 'inner', on = 'date')
            df_deficit.loc[df_deficit['Snow'] > kwargs['snow_frac'], 'ET'] = 0
        except:
            raise Exception("Snow correction can't be applied. Either no snow data presented or snow_band or asset wrong. Given snow_band: {}, snow_asset: {}".format(kwargs['snow_band'], kwargs['snow_asset']))
   
    # Calculate A and D
    df_deficit['A'] = df_deficit['ET'] - df_deficit['P']
    df_deficit['D'] = 0
    for _i in range(df_deficit.shape[0]-1):
        df_deficit.loc[_i+1, 'D'] = max((df_deficit.loc[_i+1, 'A'] + df_deficit.loc[_i, 'D']), 0)
    
    # Calculate wateryear deficit (D(t)_wy)) and merge columns with df_deficit
    df_deficit_wy = pd.DataFrame()
    for wy in df_deficit.wateryear.unique():
        temp = df_deficit[df_deficit['wateryear'] == wy][['date','ET','P']]
        temp['A'] = temp['ET'] - temp['P']
        temp['D_wy'] = 0
        temp = temp.reset_index()
        for _i in range(temp.shape[0]-1):
            temp.loc[_i+1, 'D_wy'] = max((temp.loc[_i+1, 'A'] + temp.loc[_i, 'D_wy']), 0)
        df_deficit_wy = pd.concat([df_deficit_wy, temp])
    df_deficit_wy = df_deficit_wy[['date','D_wy']]
    df_deficit = df_deficit.merge(df_deficit_wy, how = 'left', on = 'date')
    #self.deficit_timeseries = df_deficit
    #self.smax = round(df_deficit.D.max())
    #self.maxdmax = round(df_deficit.D_wy.max())
    return df_deficit


def deficit_bursts(df):
    """
    Still under development!! Get a dataframe with the length and maximum deficit of each "burst" (i.e. deficits that are continuously above zero).

    Args:
        df (:obj:`df`): dataframe with columns for date and D (for deficit), at minimum.

    Returns: 
        :obj:`df`: dataframe with columns for start and end dates, and max_D [mm], duration [days] and start and end wateryear of each burst.
    """
    # get all places where deficit = 0
    df_zero = df[df.D == 0]

    # get the start and end dates of the bursts and add in an artificial 'burst' starting on the first date of the year and ending at the start of the next burst
    start_of_bursts = []
    end_of_bursts = []
    max_D = []

    # For all the dates where the deficit = 0, find the time between the dates and the maximum value between the dates
    for j in range(len(df_zero['date'])-1):
        if (df_zero['date'][j+1] > df_zero['date'][j]):
            mask2 = (pd.to_datetime(df['date']) >= df_zero['date'][j]) & (pd.to_datetime(df['date']) <= df_zero['date'][j+1])
            bursts = pd.DataFrame()
            bursts = df.loc[mask2]
            if (len(bursts['D'])>2):
                start_of_bursts.append(bursts['date'][0])
                end_of_bursts.append(bursts['date'][-1])
                max_D.append(bursts['D'].max())
            else: continue
        else: continue
        
    # Save in df
    burstdf = pd.DataFrame()
    burstdf['start'] = start_of_bursts
    burstdf['end'] = end_of_bursts
    burstdf['max_D'] = max_D
  
    # Clean up and add more columns
    burstdf['start'] = pd.to_datetime(burstdf['start'])
    burstdf['end'] = pd.to_datetime(burstdf['end'])
    burstdf['duration'] = (burstdf['end'] - burstdf['start']).dt.days.astype('int16')
    burstdf = burstdf.set_index(burstdf['start'])
    burstdf['start_wateryear'] = np.where(~burstdf.index.month.isin([10,11,12]),burstdf.index.year,burstdf.index.year+1)
    burstdf = burstdf.reset_index(drop=True)
    burstdf = burstdf.set_index(burstdf['end'])
    burstdf['end_wateryear'] = np.where(~burstdf.index.month.isin([10,11,12]),burstdf.index.year,burstdf.index.year+1)
    burstdf = burstdf.reset_index(drop=True)

    return burstdf


"""
def getRA(site_lat_decimal):
  ## J (julian day)
  J = (np.linspace(1,365,365))
  ## Gsc (solar constant)  [MJ m-2 day -1] 
  Gsc = 0.0820
  ## inverse relative distance Earth-Sun
  dr = 1+0.033*np.cos(((2*np.pi)/365)*J) 
  ## delta = solar declination [rad]
  delta = 0.409*np.sin(((2*np.pi)/365)*J-1.39)  
  ## psi [rad] = convert decimal latitude to radians
  psi = (np.pi/180)*(32.3863) 
  ## omega_s (ws)= solar time angle at beginning of period [rad]
  omega_s = np.arccos(-np.tan(psi)*np.tan(delta)) 
  ## [ws * sin(psi) * sin(delta) + cos(psi) * cos(delta) * sin(ws)]
  angles = omega_s * np.sin(psi) * np.sin(delta) + np.cos(psi) * np.cos(delta) * np.sin(omega_s)
  RA = ((24*60)/np.pi) * Gsc * dr * angles

  df = pd.DataFrame()
  df['RA'] = RA
  df['J'] = J
  df['J']= df['J'].astype(int)
  return df


def calculate_PET(daily_data_df, latitude):#gage, p, tmin = 'tmin', tmax = 'tmax', tmean = 'tmean'):
    if 'tmax' and 'tmin' not in daily_data_df['band'].unique():
        sys.exit("'PET cannot be calsculated because PRISM temperature was not extracted.'")  
    else:
        # Get temperatures from prism
        prism_df = daily_data_df[daily_data_df['asset_name'] == 'prism']
        pet_df = pd.DataFrame()
        pet_df['tmax'] = prism_df[prism_df['band'] == 'tmax']['value'].values
        pet_df['tmin'] = prism_df[prism_df['band'] == 'tmin']['value'].values
        pet_df['date'] = pd.to_datetime(prism_df[prism_df['band'] == 'tmin']['date']).values

        # get RA (extraterrestrial radiation) with julian day column
        RA_df = getRA(latitude)
        pet_df['J'] = pd.to_datetime(pet_df['date']).dt.strftime('%j')
        pet_df['J'] = pet_df['J'].astype(int)
        
        # merge prism df with RA
        pet_df = pet_df.merge(RA_df, how='left', on=['J'])

        # calculate PET (Hargreave and Semani 1985)
        # Ep = 0.0023 · (Tmean + 17.8)·(Tmax - Tmin)^0.5 · 0.408 · Rext
        Krs = 0.0023 # Erica changed this from 0.00023 on Oct 15, 2021
        pet_df['PET'] = Krs * pet_df['RA'] * np.sqrt(pet_df['tmax'] - pet_df['tmin']) * (((pet_df['tmin']+pet_df['tmax'])/2) + 17.8)
    
        return pet_df
"""      


import pathlib
from setuptools import setup, find_packages

# The directory containing this file
HERE = pathlib.Path(__file__).parent

# The text of the README file
README = (HERE / "README.rst").read_text()

setup(
    name="waterpyk",
    version="0.1.0",
    description= "Extract, analyze, and plot hydrological timeseries for a site or watershed using the Google Earth Engine and USGS APIs",
    long_description=README,
    long_description_content_type="text/x-rst",
    url= "https://github.com/erica-mccormick/waterpyk",
    author="Erica L. McCormick",
    author_email= "erica.elmstead@gmail.com",
    license="MIT",
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
        "Intended Audience :: Science/Research",
        "Natural Language :: English",
        "Topic :: Scientific/Engineering :: GIS",
        "Development Status :: 4 - Beta"
    ],
    packages=find_packages(),
    include_package_data= True,
    package_data={'': ['layers_data/*.csv']}
)


import matplotlib
import matplotlib.pyplot as plt
import seaborn as sns
import pandas as pd
import numpy as np
from scipy import stats

default_kwarg_dictionary = {
        'plot_PET': False,
        'plot_Q': False,
        'plot_P': True,
        'plot_D': True,
        'plot_Dwy': True,
        'plot_ET': False,
        'plot_ET_dry': False,
        'color_PET': 'red',
        'color_Q': 'blue',
        'color_P': '#b1d6f0',
        'color_D': 'black',
        'color_Dwy':'black',
        'color_ET': 'purple',
        'color_WY': False,
        'markeredgecolor': 'black',
        'linestyle_PET':'-',
        'linestyle_Q':'-',
        'linestyle_P':'-',
        'linestyle_D': '-',
        'linestyle_Dwy':'--',
        'linestyle_ET': '-',
        'lw': 1.5, 
        'xmin': '2003-10-01',
        'xmax': '2020-10-01',
        'legend': True,
        'dpi': 300,
        'figsize': (6,4),
        'xlabel': 'Date',
        'ylabel': '[mm]',
        'twinx': False,
        'title': None
        }

def plot_p_distribution(studyareaobject = None, df_wateryear_totals = None, smax = None, **plot_kwargs):
    """Plot distribution of wateryear precipitation (P) and root-zone water storage capacity (Smax) in the form of
    Rempe-McCormick et al. (in prep) Fig 1c,f. 
    Either studyarea object must be supplied OR the df_wateryear_totals and smax param with the necessary data for plotting.
    
    Args:
        studyareaobject (object from studyarea class, optional):
        df_wateryear_totals (:obj:`df`, optional): dataframe with wateryear cumulative P and a column named 'P' at minimum.
        smax (int or float, optional): value for root-zone water storage capacity (Smax) in mm
        **xmin (int, optional): default = 0
        **xmax (int, optional): default = 4000
        **xlabel (str, optional): default = 'mm'
        **ylabel (str, optional): default = 'Density'
        **dpi (int, optional): default = 300
        **figsize (tuple, optional): default = (6,4)
        **legend (bool, optional): default = True
        **title (str, optional): default = None
        
    Returns:
        :obj:`fig`
    """
    
    print('Plotting precip distribution...')
    updated_kwargs = {'xmin': 0, 'xmax':4000, 'xlabel':'mm', 'ylabel':'Density'}
    default_kwargs = {**default_kwarg_dictionary, **updated_kwargs}    
    plot_kwargs = {**default_kwargs, **plot_kwargs}    
    
    if studyareaobject is not None:
        wateryear_totals = studyareaobject.wateryear_totals
        smax = studyareaobject.smax
    elif df_wateryear_totals is not None:
        wateryear_totals = df_wateryear_totals
        smax = smax
    fig, ax = plt.subplots(dpi=plot_kwargs['dpi'], figsize = plot_kwargs['figsize'])
    sns.kdeplot(data=wateryear_totals, x="P", color = '#79C3DB', alpha = 0.3)
    x = ax.lines[-1].get_xdata()
    y = ax.lines[-1].get_ydata()
    ax.fill_between(x, 0, y, color='#79C3DB', alpha=0.3, label=r'$\mathrm{P}_{wy (2003-2020)}\/\mathrm{(mm)}$')
    ax.fill_between(x, 0, y, where=x<smax, color='#ED9935', hatch = 'x', alpha=0.4)
    plt.axvspan(smax, smax, alpha=1, color='#ED9935', label = r'$\mathrm{S}_{max}\/\mathrm{(mm)}$')
    ax.set_xlim(plot_kwargs['xmin'], plot_kwargs['xmax'])
    if plot_kwargs['title'] is not None: fig.suptitle(plot_kwargs['title'])
    if plot_kwargs['legend']: ax.legend(loc = 'best')
    ax.set_xlabel(plot_kwargs['xlabel'])
    ax.set_ylabel(plot_kwargs['ylabel'])
    return fig

def plot_timeseries(studyareaobject = None, df_daily_wide = None, deficit_timeseries = None, **plot_kwargs):
    """Plot daily (wateryear cumulative) timeseries, following Rempe-McCormick et al. (in prep) Fig 1b,e. 
    Either studyarea object must be supplied OR the df_daily_wide and df_timeseries params with the necessary data for plotting.
    Use **kwargs to choose which datasets to plot, from the options of P, ET, deficit, deficit_wateryear, and streamflow (Q) if available.
    
    Args:
        studyareaobject (object from studyarea class, optional):
        df_daily_wide (:obj:`df`, optional): dataframe with daily cumulative wateryear timeseries and columns corresponding to the chosen data (from **kwargs), at minimum, such as 'ET_cumulative', 'P_cumulative', and 'Q_mm_cumulative', and 'date' columns.
        deficit_timeseries (:obj:`df`, optional): dataframe with deficit timeseries and columns 'D', 'D_wy' and 'date' at minimum.
        **xmin (str, optional): default = '2003-10-01'
        **xmax (str, optional): default = '2020-10-01'
        **xlabel (str, optional): default = 'Date'
        **ylabel (str, optional): default = '[mm]'
        **dpi (int, optional): default = 300
        **figsize (tuple, optional): default = (6,4)
        **legend (bool, optional): default = True
        **title (str, optional): default = None
        **plot_Q (bool, optional): default = False
        **plot_P (bool, optional): default = True
        **plot_D (bool, optional): default = True
        **plot_Dwy (bool, optional): default = True
        **plot_ET (bool, optional): default = False
        **plot_ET_dry (bool, optional): default = False
        **color_Q (str, optional): default = 'blue'
        **color_P (str, optional): default = '#b1d6f0'
        **color_D (str, optional): default = 'black'
        **color_Dwy (str, optional): default = 'black'
        **color_ET (str, optional): default = 'purple'
        **linestyle_Q (str, optional): default = '-'
        **linestyle_P (str, optional): default = '-'
        **linestyle_D (str, optional): default = '-'
        **linestyle_Dwy (str, optional): default = '--'
        **linestyle_ET (str, optional): default = '-'
        **lw (float, optional): lineweight, default = 1.5
        
    Returns:
        :obj:`fig`
    """
    
    print('Plotting timeseries...')
    plot_kwargs = {**default_kwarg_dictionary, **plot_kwargs}    
    
    if studyareaobject is not None:
        df_wy = studyareaobject.daily_df_wide
        df_d = studyareaobject.deficit_timeseries
    elif df_daily_wide is not None:
        df_wy = df_daily_wide
        df_d = deficit_timeseries
    else:
        print('No data was specified for plotting')
     
    df_wy['date'] = pd.to_datetime(df_wy['date'])
    df_d['date'] = pd.to_datetime(df_d['date'])   
    
    fig, ax = plt.subplots(dpi=plot_kwargs['dpi'], figsize = plot_kwargs['figsize'])
      
    # Plot things if given in kwargs
    if plot_kwargs['plot_Q']:
        if studyareaobject.kind == 'watershed':
            ax.plot(df_wy['date'], df_wy['Q_mm_cumulative'], plot_kwargs['linestyle_Q'], color=plot_kwargs['color_Q'], lw = plot_kwargs['lw'], label= 'Q (mm)')
    if plot_kwargs['plot_PET']:
        ax.plot(df_wy['date'], df_wy['PET_cumulative'], plot_kwargs['linestyle_PET'], color=plot_kwargs['color_PET'], lw = plot_kwargs['lw'], label= 'PET (mm)') 
    if plot_kwargs['plot_P']:
        ax.fill_between(df_wy['date'], 0, df_wy['P_cumulative'],color='#b1d6f0', label='P (mm)', alpha = 0.7)
    if plot_kwargs['plot_ET']:
        ax.plot(df_wy['date'], df_wy['ET_cumulative'], plot_kwargs['linestyle_ET'], color=plot_kwargs['color_ET'], lw = plot_kwargs['lw'], label= 'ET (mm)')
    if plot_kwargs['plot_D']:
        ax.plot(df_d['date'], df_d['D'], plot_kwargs['linestyle_D'], color=plot_kwargs['color_D'], lw = plot_kwargs['lw'], label=r'$\mathrm{D(t)}\/\mathrm{(mm)}$')
    if plot_kwargs['plot_Dwy']:
        ax.plot(df_d['date'], df_d['D_wy'], plot_kwargs['linestyle_Dwy'], color=plot_kwargs['color_Dwy'], lw = plot_kwargs['lw'], label=r'$\mathrm{D}_{wy}\/\mathrm{(mm)}$')
    ax.set_xlim(pd.to_datetime(plot_kwargs['xmin'], exact = False), pd.to_datetime(plot_kwargs['xmax'], exact = False))
    if plot_kwargs['title'] is not None: ax.set_title(plot_kwargs['title'])
    if plot_kwargs['legend'] == True: ax.legend(loc = 'best')
    ax.set_xlabel(plot_kwargs['xlabel'])
    ax.set_ylabel(plot_kwargs['ylabel'])
    return fig


def plot_wateryear_totals(studyareaobject = None, df_wateryear_totals = None, **plot_kwargs):
        
    """Plot summed wateryear timeseries. 
    Either studyarea object must be supplied OR the df_wateryear_totals param with the necessary data for plotting.
    Use **kwargs to choose which datasets to plot, from the options of P, ET, ET_summer, deficit, and streamflow (Q) if available.

    Args:
        studyareaobject (object from studyarea class, optional):
        df_wateryear_totals (:obj:`df`, optional): dataframe with cumulative wateryear timeseries and columns corresponding to the chosen data (from **kwargs), at minimum, such as 'P', 'D', 'ET', 'ET_summer', and 'Q_mm', and 'wateryear' columns.
        **xmin (int, optional): default = 2003
        **xmax (int, optional): default = 2020
        **xlabel (str, optional): default = 'Wateryear'
        **ylabel (str, optional): default = '[mm]'
        **dpi (int, optional): default = 300
        **figsize (tuple, optional): default = (6,4)
        **legend (bool, optional): default = True
        **title (str, optional): default = None
        **plot_Q (bool, optional): default = True
        **plot_P (bool, optional): default = True
        **plot_D (bool, optional): default = True
        **plot_ET (bool, optional): default = True
        **plot_ET_dry (bool, optional): default = False
        **color_Q (str, optional): default = 'blue'
        **color_P (str, optional): default = '#b1d6f0'
        **color_D (str, optional): default = 'black'
        **color_ET (str, optional): default = 'purple'
        **linestyle_Q (str, optional): default = '-o'
        **linestyle_P (str, optional): default = '-o'
        **linestyle_D (str, optional): default = '-'
        **linestyle_ET (str, optional): default = '-o'
        **lw (float, optional): lineweight, default = 1.5
        **twinx (bool, optional): Put ET data on twin x-axis to P data. default = True
        
    Returns:
        :obj:`fig`
    """
    print('Plotting wateryear data...')
    if studyareaobject is not None:
        df = studyareaobject.wateryear_totals
    elif df_wateryear_totals is not None:
        df = df_wateryear_totals
    else:
        print('No data was specified for plotting.')
    # Get kwargs set up
    updated_kwargs = {
            'plot_Q': True,
            'plot_P': True,
            'plot_ET': True,
            'plot_D':True,
            'twinx':True,
            'xmin':2003,
            'xmax':2020,
            'linestyle_Q':'-o',
            'linestyle_ET':'-o',
            'linestyle_P':'-o',
            'xlabel':'Wateryear'
        }
    
    default_kwargs = {**default_kwarg_dictionary, **updated_kwargs}    
    plot_kwargs = {**default_kwargs, **plot_kwargs}    
        
    fig, ax = plt.subplots(dpi=plot_kwargs['dpi'], figsize = plot_kwargs['figsize'])
    if plot_kwargs['plot_PET']:
        print('Plotting PET!')
        ax.plot(df['wateryear'], df['PET'], plot_kwargs['linestyle_PET'], color = plot_kwargs['color_PET'], lw = plot_kwargs['lw'], markeredgecolor = plot_kwargs['markeredgecolor'], label = r'$\mathrm{PET}_{wy}\/\mathrm{(mm)}$')
    if plot_kwargs['plot_P']:
        ax.plot(df['wateryear'], df['P'], plot_kwargs['linestyle_P'], color = plot_kwargs['color_P'], lw = plot_kwargs['lw'], markeredgecolor = plot_kwargs['markeredgecolor'], label = r'$\mathrm{P}_{wy}\/\mathrm{(mm)}$')
    if plot_kwargs['plot_Q']:
        if studyareaobject.kind == 'watershed':
            ax.plot(df['wateryear'], df['Q_mm'], plot_kwargs['linestyle_Q'], color = plot_kwargs['color_Q'], lw = plot_kwargs['lw'], markeredgecolor = plot_kwargs['markeredgecolor'], label = r'$\mathrm{Q}_{wy}\/\mathrm{(mm)}$')
    if plot_kwargs['twinx']:
        ax2 = ax.twinx()
        ax2.set_ylabel('ET (mm)', color = plot_kwargs['color_ET'])
    else: ax2 = ax
    if plot_kwargs['plot_ET']:
        ax2.plot(df['wateryear'], df['ET'], plot_kwargs['linestyle_ET'], color = plot_kwargs['color_ET'], lw = plot_kwargs['lw'], markeredgecolor = plot_kwargs['markeredgecolor'], label = r'$\mathrm{ET}_{wy}\/\mathrm{(mm)}$')
    if plot_kwargs['plot_ET_dry']:
        ax2.plot(df['wateryear'], df['ET_summer'], ':o', color = plot_kwargs['color_ET'], lw = plot_kwargs['lw'], markeredgecolor = plot_kwargs['markeredgecolor'], label = r'$\mathrm{ET}_{dry}\/\mathrm{(mm)}$')
    ax.set_xlim(plot_kwargs['xmin'], plot_kwargs['xmax'])
    if plot_kwargs['title'] is not None: ax.set_title(plot_kwargs['title'])
    if plot_kwargs['legend']: ax.legend(loc = 'best')
    ax.set_xlabel(plot_kwargs['xlabel'])
    ax.set_ylabel(plot_kwargs['ylabel'])
    return fig



def plot_spearman(studyareaobject = None, df_wateryear_totals = None, **plot_kwargs):
    """Scatter plot of wateryear precipitation and summer (ie dry season) ET, following Rempe-McCormick et al. (in prep) Fig 1b,e.
    Spearman correlation coefficient (rho) and p-value is given in upper right corner of plot. 
    Either studyarea object must be supplied OR the df_wateryear_totals param with the necessary data for plotting (ie wateryear, P, and ET_summer columns).
    
    Args:
        studyareaobject (object from studyarea class, optional):
        df_wateryear_totals (:obj:`df`, optional): dataframe with columns 'P', 'ET_summer', and 'wateryear'
        deficit_timeseries (:obj:`df`, optional): dataframe with deficit timeseries and columns 'D', 'D_wy' and 'date' at minimum.
        **xmin (int, optional): default = 0
        **xmax (int, optional): default = 3000
        **dpi (int, optional): default = 300
        **figsize (tuple, optional): default = (6,4)
        **color_WY (bool, optional): This puts a wateryear legend and colors the dots by wateryear. default = True
        **title (str, optional): default = None
        **markeredgecolor (str, optional): default = 'black'
        **lw (float, optional): lineweight, default = 1.5
        
    Returns:
        :obj:`fig`
    """
    
    print('Plotting spearman...')
    if studyareaobject is not None:
        df = studyareaobject.wateryear_totals
    elif df_wateryear_totals is not None:
        df = df_wateryear_totals
    else:
        print('No data was specified for plotting')

    # Get kwargs set up
    updated_kwargs = {'xmin':0, 'xmax':3000, 'color_WY': True}
    default_kwargs = {**default_kwarg_dictionary, **updated_kwargs}    
    plot_kwargs = {**default_kwargs, **plot_kwargs}     

    fig, ax = plt.subplots(dpi=plot_kwargs['dpi'], figsize = plot_kwargs['figsize'])

    plot_kwargs['xlabel'] = r'$\mathrm{P}_{wy}\/\mathrm{(mm)}$'
    plot_kwargs['ylabel'] = r'$\mathrm{ET}_{dry}\/\mathrm{(mm)}$'
    plot_kwargs['legend'] = False
    ax.set_ylim(0,600)
    corr, p = stats.spearmanr(df[['P','ET_summer']])
    ax.annotate(r'$\rho$' + ' = ' + str(round(corr,2)) + '\n p-val = ' + str(round(p,4)),
                xy=(.88, .85), xycoords='figure fraction',
                horizontalalignment='right', verticalalignment='top',
                fontsize=10)
    ax.set_xlim(plot_kwargs['xmin'], plot_kwargs['xmax'])
    ax.set_xlabel(plot_kwargs['xlabel'])
    ax.set_ylabel(plot_kwargs['ylabel'])
    if plot_kwargs['title'] is not None: ax.set_title(plot_kwargs['title'])
    if plot_kwargs['legend']: ax.legend(loc = 'best')
    ax.set_xlabel(plot_kwargs['xlabel'])
    ax.set_ylabel(plot_kwargs['ylabel'])
    if plot_kwargs['color_WY']:
        cax = fig.add_axes([0.14, 0.15, 0.05, 0.70]) #[left, bottom, width, height]
        my_cmap = plt.get_cmap('tab20', 17)
        im = ax.scatter(df['P'], df['ET_summer'], c = df.wateryear, vmin=2004, vmax=2021, cmap = my_cmap, s = 150, edgecolor = 'black')
        fig.colorbar(im, cax=cax, orientation='vertical', spacing='proportional')
    else:
        ax.plot(df['P'], df['ET_summer'], 'o', color = '#a4a5ab', markersize = 12, lw = plot_kwargs['lw'],  markeredgecolor = plot_kwargs['markeredgecolor'], label = '')
    return fig


def plot_RWS(studyareaobject = None, df_deficit =  None, smax = None, **plot_kwargs):
    """Plot root-zone water storage (RWS) timeseries, following Rempe-McCormick et al. (in prep) SI Fig 1. 
    Either studyarea object must be supplied OR the df_deficit and smax params with the necessary data for plotting (ie columns for 'D' and 'date' and smax value)
    
    Args:
        studyareaobject (object from studyarea class, optional):
        deficit_timeseries (:obj:`df`, optional): dataframe with deficit timeseries and columns 'D', and 'date' at minimum.
        smax (int or float, optional): root-zone water storage capacity (Smax) in mm
        **xmin (str, optional): default = '2003-10-01'
        **xmax (str, optional): default = '2020-10-01'
        **xlabel (str, optional): default = 'Date'
        **ylabel (str, optional): default = 'RWS (mm)'
        **dpi (int, optional): default = 300
        **figsize (tuple, optional): default = (6,4)
        **legend (bool, optional): default = False
        **title (str, optional): default = None
        **color_D (str, optional): default = 'black'
        **linestyle_D (str, optional): default = '-'
        **lw (float, optional): lineweight, default = 1
        
    Returns:
        :obj:`fig`
    """
    
    print('Plotting RWS...')
    updated_kwargs = {'legend': False, 'lw':1, 'ylabel': 'RWS (mm)'}
    default_kwargs = {**default_kwarg_dictionary, **updated_kwargs}    
    plot_kwargs = {**default_kwargs, **plot_kwargs}    
    
    if studyareaobject is not None:
        df_d = studyareaobject.deficit_timeseries
        smax = studyareaobject.smax
    elif df_deficit is not None:
        df_d = df_deficit
        smax = df_deficit.max()
    
    df_d['RWS'] = df_d['D'] - smax
    df_d['RWS'] = df_d['RWS'] * -1
    df_d['date'] = pd.to_datetime(df_d['date'])

    fig, ax = plt.subplots(dpi=plot_kwargs['dpi'], figsize = plot_kwargs['figsize'])

    ax.plot(df_d['date'], df_d['RWS'], plot_kwargs['linestyle_D'], color=plot_kwargs['color_D'], lw = plot_kwargs['lw'], label=r'$\mathrm{RWS(t)}\/\mathrm{(mm)}$')
    ax.set_xlim(pd.to_datetime(plot_kwargs['xmin'], exact = False), pd.to_datetime(plot_kwargs['xmax'], exact = False))
    if plot_kwargs['title'] is not None: ax.set_title(plot_kwargs['title'])
    if plot_kwargs['legend']: ax.legend(loc = 'best')
    ax.set_xlabel(plot_kwargs['xlabel'])
    ax.set_ylabel(plot_kwargs['ylabel'])
    return fig
        
        
